# -*- coding: utf-8 -*-
"""
Code for sampling from WSI

"""

import itertools
import numpy as np

from .tissue_segmentation import roi_binary_mask
from .utils import (find_square, get_size, get_whole_image, pj_slice,
                    get_x_y, get_x_y_from_0, mask_percentage, open_image)


def sample_patch_from_wsi(slide, mask=None, mask_level=None, 
                          patch_size=(512, 512), analyse_level=0,
                          list_func=[]):
    """
    Samples one tile from a slide where mask is 1

    Args:
        slide : String or open_slide object. The slide from which you wish to sample.
        mask : None by default or binary numpy array, where positive pixels correspond to tissue area and
               negative pixels to background areas in the slide. 
        mask_level : Integer or None. Level to which apply mask_function to the rgb 
                     image of the slide at that resolution. mask_function(slide[mask_level])
                     will return the binary image corresponding to the tissue.
        patch_size : Tuple of integers or None. If none the default tile size will (512, 512).
        analyse_level : Integer. Level resolution to use for extracting the tiles.
        list_func : None or list of functions to apply to the tiles. Useful to filter the tiles
                    that are part of the tissue mask. Very useful if the tissue mask is bad and 
                    samples many white background tiles, in this case it is interesting to add 
                    a function to eliminate tiles that are too white, like the function white_percentage.
    Returns:
        A list of 5 parameters corresponding to: [x, y, size_x_level, size_y_level, level]
    """       
    slide = open_image(slide)
    if mask_level is None:
        mask_level = slide.level_count - 1
    slide_png = get_whole_image(slide, mask_level, numpy=True)
    size_l = get_size(slide, patch_size, analyse_level, mask_level)
    size_l = np.array(size_l)
    if mask is None:
        mask = np.ones_like(slide_png)[:, :, 0]
    x_mask, y_mask = np.where(mask)
    indice = np.random.randint(len(x_mask))
    point = np.array([x_mask[indice], y_mask[indice]])
    sub_img = pj_slice(slide_png, point - size_l // 2, point + size_l // 2)
    criterias = []
    for function in list_func:
        criterias.append(function(sub_img))
    if all(criterias):
        para = find_square(slide, point, mask_level, analyse_level, patch_size)
    else:
        para = None
    return para


def remove_sample_from_mask(slide, para, mask, mask_level):
    """
    Given a square patch and a mask, removes the patch from mask.
    So that it can't be choosen again for instance...
    Args:
        slide : String or open_slide object. The slide from which you wish to sample.
        para : A list of 5 integer parameters corresponding to: [x, y, size_x_level, size_y_level, level]
        mask : None by default or binary numpy array, where positive pixels correspond to tissue area and
               negative pixels to background areas in the slide. 
        mask_level : Integer or None. Level to which apply mask_function to the rgb 
                     image of the slide at that resolution. mask_function(slide[mask_level])
                     will return the binary image corresponding to the tissue.    Returns: 
        An updated version of mask where the tile given by para and mask_level
        is removed.
    """
    if para is not None:
        point_0 = (para[1], para[0])
        size_l = (para[2], para[3])
        analyse_level = para[4]
        point_mask_res = np.array(get_x_y_from_0(slide, point_0, mask_level))
        size_mask_res = np.array(get_size(slide, size_l, analyse_level, mask_level))
        start_point = np.array([point_mask_res - size_mask_res // 2, (0, 0)]).max(axis=0)
        end_point = start_point + size_mask_res
        mask[start_point[0]:end_point[0], start_point[1]:end_point[1]] = 0

    return mask


def random_wsi_sampling(n_samples, slide, mask=None,
                        mask_level=None, patch_size=(512, 512),
                        analyse_level=0, with_replacement=False,
                        list_func=[]):
    """
    Randomly generate patches from slide.

    Args:
        n_samples : Integer, number of tiles to extract from the slide with the 
                    sampling method "random_sampling".
        slide : String or open_slide object. The slide from which you wish to sample.
        mask : None by default or binary numpy array, where positive pixels correspond to tissue area and
               negative pixels to background areas in the slide. 
        mask_level : Integer or None. Level to which apply mask_function to the rgb 
                     image of the slide at that resolution. mask_function(slide[mask_level])
                     will return the binary image corresponding to the tissue.
        patch_size : Tuple of integers or None. If none the default tile size will (512, 512).
        analyse_level : Integer. Level resolution to use for extracting the tiles.
        with_replacement : Bool, default to False. Wether or not you can sample with replacement. 
                           Here, if True, we would remove the previous patches from the original 
                           mask at each iteration.
        list_func : None or list of functions to apply to the tiles. Useful to filter the tiles
                    that are part of the tissue mask. Very useful if the tissue mask is bad and 
                    samples many white background tiles, in this case it is interesting to add 
                    a function to eliminate tiles that are too white, like the function white_percentage.

    Returns:
        A list of 5 parameters corresponding to: [x, y, size_x_level, size_y_level, level]
    """    
    list_para = []
    if mask_level is None:
        mask_level = slide.level_count - 1
    slide_png = get_whole_image(slide, mask_level, numpy=True)
    if mask is None:
        mask = np.ones_like(slide_png)[:, :, 0]
    mask = mask.astype('bool')

    for _ in range(n_samples):
        para = sample_patch_from_wsi(slide, mask, mask_level, patch_size,
                                     analyse_level, list_func)
        if para is not None:
            list_para.append(para)
        if not with_replacement:
            mask = remove_sample_from_mask(slide, para, mask, mask_level)
            if mask.sum() == 0:
                break
    return list_para


def grid_blob(slide, point_start, point_end, patch_size,
              analyse_level):
    """
    Forms a uniform grid starting from the top left point point_start
    and finishes at point point_end of size patch_size at level analyse_level
    for the given slide.
    Args:
        slide : String or open_slide object. 
        point_start : Tuple like object of integers of size 2.
        point_end : Tuple like object of integers of size 2.
        patch_size : Tuple like object of integers of size 2.
        analse_level : Integer. Level resolution to use for extracting the tiles.
    Returns:
        List of coordinates of grid.
    """
    if analyse_level == 0:
        patch_size_0 = patch_size
    else:
        patch_size_0 = get_size(slide, patch_size, analyse_level, 0)
    size_x, size_y = patch_size_0
    list_x = range(point_start[0], point_end[0], size_x)
    list_y = range(point_start[1], point_end[1], size_y)
    return list(itertools.product(list_x, list_y))


def correct_patch(coord, slide, analyse_level, patch_size):
    """
    Correct patch by shifting so that the whole square patch can fit.
    Args:
        coord : A tuple like object of size 2 with integers.
        slide : String or open_slide object.
        analyse_level : Integer. Level resolution to use for extracting the tiles.
        patch_size : Tuple of integers or None.
    Returns:
        A numpy array of size 2 corresponding to the corrected or not
        orignal coord.
    """
    max_dimensions = np.array(slide.dimensions)[::-1]

    if analyse_level != 0:
        patch_size_0 = get_size(slide, patch_size, analyse_level, 0)
    else:
        patch_size_0 = patch_size
    coord = np.array([coord, max_dimensions - 1 - patch_size_0]).min(axis=0)
    return coord

def check_patch(slide, mask, coord_grid_0, mask_level, 
                patch_size, analyse_level,
                list_func, mask_tolerance=0.5,
                allow_overlapping=False,
                margin=0):
    """
    Filters a list of possible coordinates with a set of filtering parameters.

    Args:
        slide : String or open_slide object. The slide from which you wish to sample.
        mask : Binary numpy array, where positive pixels correspond to tissue area and
               negative pixels to background areas in the slide.
        coord_grid_0 : List of list of two elements where each (sub) list can be described as 
                       possible coordinates for a possible tile at analyse_level.
        mask_level : Integer or None. Level to which apply mask_function to the rgb 
                     image of the slide at that resolution. mask_function(slide[mask_level])
                     will return the binary image corresponding to the tissue.
        patch_size : Tuple of integers or None. If none the default tile size will (512 + margin, 512 + margin).
        analyse_level : Integer. Level resolution to use for extracting the tiles.
        list_func : None or list of functions to apply to the tiles. Useful to filter the tiles
                    that are part of the tissue mask. Very useful if the tissue mask is bad and 
                    samples many white background tiles, in this case it is interesting to add 
                    a function to eliminate tiles that are too white, like the function white_percentage.
        mask_tolerance : Float between 0 and 1. A tile will be accepted if pixel_in_mask / total_pixel > value.
                         So if mask_tolerance = 1, only tiles that are completly within the mask are accepted.
        allow_overlapping : Bool. False by default, this parameter does not influence the 'overlapping' parameter
                            above. This parameter only influences the tiles on that reach the border. In particular
                            if the tile extends out of the boundaries of the slide, in this case if allow_overlapping
                            is set to true, it will correct the extracted tile by allowing it to overlappe with it's
                            neighbouring tile.
        margin : Integer. By default set to 0, number of pixels at resolution 0 to add
                 to patch_size on each side. (different to overlapping as this is at resolution 0)
    Returns:
        List of parameters where each parameter is a list of 5 elements
        [x, y, size_x_level, size_y_level, level]
    """
    slide_png = get_whole_image(slide, level=mask_level, numpy=True)
    assert slide_png.shape[0:2] == mask.shape[0:2], "Raise value, mask not of the right shape {}".format(mask.shape[0:2])
    shape_mask = np.array(mask.shape[0:2])
    parameters = []
    patch_size_l = get_size(slide, patch_size, analyse_level, mask_level)
    radius = np.array([max(el // 2, 1) for el in patch_size_l])
    for coord_0 in coord_grid_0:
        coord_l = get_x_y_from_0(slide, coord_0, mask_level)
        # coord_0 = np.array(coord_0)[::-1]
        # coord_l = np.array(coord_l)[::-1]
        point_cent_l = [coord_l + radius, shape_mask - 1 - radius]
        point_cent_l = np.array(point_cent_l).min(axis=0)
        if mask_percentage(mask, point_cent_l, radius, mask_tolerance): ## only checking center point
            criterias = []
            sub_img = pj_slice(slide_png, point_cent_l - radius, point_cent_l + radius)
            for function in list_func:
                criterias.append(function(sub_img))
            if all(criterias):
                still_add = True
                if ((coord_l + radius) != point_cent_l).any():
                    # If the patch is going off the border
                    still_add = False
                    if allow_overlapping:
                        coord_0 = correct_patch(coord_0, slide, analyse_level, patch_size)
                        still_add = True
                if still_add:
                    sub_param = [coord_0[1] - margin, coord_0[0] - margin, \
                                 patch_size[0] + 2 * margin, patch_size[1] + 2 * margin, \
                                 analyse_level]
                    parameters.append(sub_param)
    return parameters


def patch_sampling(slide, seed=None, mask_level=None,
                   mask_function=roi_binary_mask, sampling_method=None,
                   analyse_level=0, patch_size=None, overlapping=0,
                   list_func=None, mask_tolerance=0.5, allow_overlapping=False,
                   n_samples=10, with_replacement=False):
    """
    Returns a list of tiles from slide given a mask generating method
    and a sampling method
    Args:
        slide : String or open_slide object. The slide from which you wish to sample.
        seed : Integer or None. Seed value to use for setting numpy randomness.
        mask_level : Integer or None. Level to which apply mask_function to the rgb 
                     image of the slide at that resolution. mask_function(slide[mask_level])
                     will return the binary image corresponding to the tissue.
        mask_function : Function that returns a binary image of same size as input. 
                        Mask_function is applied in order to determine the tissue areas on 
                        the slide.
        sampling_method : String. Possible values are 'grid', 'random_patches' for the 
                          patch sampling method.
                          If grid, we apply a grid on the tissue and extra all the tiles that
                          overlap on the tissue mask.
                          If random_patches, the tiles will be sampled at random from the tissue 
                          mask until no more available tissue or that n_samples has been reached.
        analyse_level : Integer. Level resolution to use for extracting the tiles.
        patch_size : Tuple of integers or None. If none the default tile size will (512, 512).
        overlapping : Integer. By default set to 0, number of pixels at analyse level to add
                      to patch_size on each side.
        list_func : None or list of functions to apply to the tiles. Useful to filter the tiles
                    that are part of the tissue mask. Very useful if the tissue mask is bad and 
                    samples many white background tiles, in this case it is interesting to add 
                    a function to eliminate tiles that are too white, like the function white_percentage.
        mask_tolerance : Float between 0 and 1. A tile will be accepted if pixel_in_mask / total_pixel > value.
                         So if mask_tolerance = 1, only tiles that are completly within the mask are accepted.
        allow_overlapping : Bool. False by default, this parameter does not influence the 'overlapping' parameter
                            above. This parameter only influences the tiles on that reach the border. In particular
                            if the tile extends out of the boundaries of the slide, in this case if allow_overlapping
                            is set to true, it will correct the extracted tile by allowing it to overlappe with it's
                            neighbouring tile. Only taken into account for the method "grid".
        n_samples : Integer, default to 10, number of tiles to extract from the slide with the 
                    sampling method "random_sampling".
        with_replacement : Bool, default to False. Wether or not you can sample with replacement in the case
                           of random sampling.

    Returns:
        List of parameters where each parameter is a list of 5 elements
        [x, y, size_x_level, size_y_level, level]
    """
    np.random.seed(seed)
    slide = open_image(slide)

    if patch_size is None:
        patch_size = (512, 512)
    if list_func is None:
        list_func = list()
    if mask_level is None:
        mask_level = slide.level_count - 1

    wsi_tissue = get_whole_image(slide, level=mask_level, numpy=True)
    wsi_mask = mask_function(wsi_tissue)

    if sampling_method == 'grid':  # grid is just grid_etienne with marge = 0
        min_row, min_col, max_row, max_col = 0, 0, *wsi_mask.shape
        point_start_l = min_row, min_col
        point_end_l = max_row, max_col
        point_start_0 = get_x_y(slide, point_start_l, mask_level)
        point_end_0 = get_x_y(slide, point_end_l, mask_level)
        grid_coord = grid_blob(slide, point_start_0, point_end_0, patch_size,
                               analyse_level)

        margin_mask_level = get_size(slide, (overlapping, 0),
                                     0, analyse_level)[0]
        parameter = check_patch(slide, wsi_mask, grid_coord,
                                mask_level, patch_size, analyse_level,
                                list_func, mask_tolerance=mask_tolerance,
                                allow_overlapping=allow_overlapping,
                                margin=margin_mask_level)
        return_list = parameter
    elif sampling_method == "random_patches":
        return_list = random_wsi_sampling(n_samples, slide, wsi_mask,
                                          mask_level, patch_size, analyse_level,
                                          with_replacement=with_replacement,
                                          list_func=list_func)
    elif sampling_method == "random_patches_with_border":
        raise NameError('sampling method random_patches_with_border is not yet implemented...')
    else:
        raise NameError('sampling method is unknown...')
    return return_list



# -*- coding: utf-8 -*-
"""
Created on Tue May 03 16:32:53 2016

@author: Peter
"""

import numpy as np
import openslide
import warnings

def load_pred_cnn():
    import sys
    import useful_wsi as module
    from os.path import join
    path = module.__path__[0]
    path = join(path, "..", "example", "best_wsi_tissue_segmentation")
    sys.path.append(path)
    from tissue_prediction import pred_cnn
    return pred_cnn

def open_image(slide):
    """
    Open openslide image
    Args:
        slide : String or other object (hopefully openslide object).
    Returns: 
        If string, returns an openslide object else return the input.
    """
    if isinstance(slide, str):
        slide = openslide.open_slide(slide)
    return slide

def get_image(slide, para, numpy=True):
    """
    Returns cropped image given a set of parameters.
    You can feed a string or an openslide image.
    Args:
        slide : String or openslide object from which we extract.
        para : List of 5 integers corresponding to: [x, y, size_x_level, size_y_level, level]
        numpy : Boolean, by default True, wether or not to convert the output to numpy array instead
                of PIL image.
    Returns:
        A tile (or crop) from slide corresponding to para. It can be a numpy array
        or a PIL image.

    """
    if len(para) != 5:
        raise NameError("Not enough parameters...")
    slide = open_image(slide)
    slide = slide.read_region((para[0], para[1]), para[4], (para[2], para[3]))

    if numpy:
        slide = np.array(slide)[:, :, 0:3]
    return slide

def get_whole_image(slide, level=None, numpy=True):
    """
    Return whole image at a certain level.
    Args:
        slide : String or openslide object from which we extract.
        level : Integer, by default None. If None the value is set to
                the maximum level minus one of the slide. Level at which
                we extract.
        numpy : Boolean, by default True, wether or not to convert the output to numpy array instead
                of PIL image.
    Returns:
        A numpy array or PIL image corresponding the whole slide at a given
        level.
    """
    if isinstance(slide, str):
        slide = openslide.open_slide(slide)
    
    if level is None:
        level = slide.level_count - 1
    elif level > slide.level_count - 1:
        print(" level ask is too low... It was setted accordingly")
        level = slide.level_count - 1
    sample = slide.read_region((0, 0), level, slide.level_dimensions[level])
    if numpy:
        sample = np.array(sample)[:, :, 0:3]
    return sample
    
def get_x_y(slide, point_l, level, integer=True):
    """
    Given a point point_l = (x_l, y_l) at a certain level. This function
    will return the coordinates associated to level 0 of this point point_0 = (x_0, y_0).
    Args:
        slide : Openslide object from which we extract.
        point_l : A tuple, or tuple like object of size 2 with integers.
        level : Integer, level of the associated point.
        integer : Boolean, by default True. Wether or not to round
                  the output.
    Returns:
        A tuple corresponding to the converted coordinates, point_0.
    """
    x_l, y_l = point_l
    size_x_l = slide.level_dimensions[level][0]
    size_y_l = slide.level_dimensions[level][1]
    size_x_0 = float(slide.level_dimensions[0][0])
    size_y_0 = float(slide.level_dimensions[0][1])
  
    x_0 = x_l * size_x_0 / size_x_l
    y_0 = y_l * size_y_0 / size_y_l
    if integer:
        point_0 = (int(x_0), int(y_0))
    else:
        point_0 = (x_0, y_0)
    return point_0

def get_x_y_from_0(slide, point_0, level, integer=True):
    """
    Given a point point_0 = (x0, y0) at level 0, this function will return 
    the coordinates associated to the level 'level' of this point point_l = (x_l, y_l).
    Inverse function of get_x_y
    Args:
        slide : Openslide object from which we extract.
        point_0 : A tuple, or tuple like object of size 2 with integers.
        level : Integer, level to convert to.  
        integer : Boolean, by default True. Wether or not to round
                  the output.
    Returns:
        A tuple corresponding to the converted coordinates, point_l.
    """
    x_0, y_0 = point_0
    size_x_l = slide.level_dimensions[level][0]
    size_y_l = slide.level_dimensions[level][1]
    size_x_0 = float(slide.level_dimensions[0][0])
    size_y_0 = float(slide.level_dimensions[0][1])
  
    x_l = x_0 * size_x_l / size_x_0
    y_l = y_0 * size_y_l / size_y_0
    if integer:
        point_l = (round(x_l), round(y_l))
    else:
        point_l = (x_l, y_l)
    return point_l
                      
def get_size(slide, size_from, level_from, level_to, integer=True):
    """
    Given a size (size_from) at a certain level (level_from), this function will return
    a new size (size_to) but at a different level (level_to).
    Args:
        slide : Openslide object from which we extract.
        size_from : A tuple, or tuple like object of size 2 with integers.
        level_from : Integer, initial level.
        level_to : Integer, final level.
        integer : Boolean, by default True. Wether or not to round
                  the output.
        Returns:
            A tuple, or tuple like object of size 2 with integers corresponding 
            to the new size at level level_to. Or size_to.
    """
    size_x, size_y = size_from
    downsamples = slide.level_downsamples
    scal = float(downsamples[level_from]) / downsamples[level_to]
    if integer:
        func_round = round
    else:
        func_round = lambda x: x
    size_x_new = func_round(float(size_x) * scal)
    size_y_new = func_round(float(size_y) * scal)
    size_to = size_x_new, size_y_new
    return size_to

def find_square(slide, point, level_from, level_to, nber_pixels):
    """
    For a given pair of coordinates at level level_from, returns a 
    square centered on these coordinates at level level_to. 
    The square will be of height h and of width w where if nber_pixel is
    integer h = w = ceil(sqrt(nber_pixel)), if nber_pixel is tuple 
    h, w = nber_pixel.
    Args:
        slide : Openslide object from which we extract.
        point : A tuple, or tuple like object of size 2 with integers corresponding
                to the center of the patch. 
        level_from : Integer, initial level.
        level_to : Integer, final level.
        nber_pixels: If integer, h = w = ceil(sqrt(nber_pixel))
                     If tuple, h, w = nber_pixel

    Returns:
        List of 5 integers corresponding to: [x, y, size_x_level, size_y_level, level]
    """
    x_0, y_0 = get_x_y(slide, point, level_from)
    if isinstance(nber_pixels, int):
        height = int(np.ceil(np.sqrt(nber_pixels)))
        size_from = (height, height)
    elif isinstance(nber_pixels, tuple):
        size_from = nber_pixels
    else:
        raise NameError("Issue number 0002, nber_pixels should be of type int or tuple")
    if level_to == 0:
        size_level_to = size_from
    else:
        size_level_to = get_size(slide, size_from, level_to, 0)

    new_x = max(x_0 - size_level_to[0] / 2, 0)
    new_y = max(y_0 - size_level_to[1] / 2, 0)
    return_list = [new_y, new_x, size_from[0], size_from[1], level_to]
    return_list = [int(el) for el in return_list]
    return return_list


def white_percentage(rgb, white_thresh=220, tolerance=0.8):
    """
    Given an rgb image, computes a white score and return 
    true or false depending if this satisfies the condition.
    Args:
        rgb : 3D matrix where the last dimension corresponds to 
              the image channels.
        white_thresh : Integer, by default set to 220. This default value
                       is only valid if rgb is of type uint8.
        tolerance : A float between 0 and 1. By default 0.5.

    Returns:
        A boolean. If true, keep, if false, discard.
    """
    score = (rgb.mean(axis=-1) > white_thresh).sum() # amount of white pixels
    score = score / (rgb.shape[0] * rgb.shape[1])
    accepted = score < tolerance
    return accepted

# def check_borders_correct(array_np, point_0, point_1):
#     shape = array_np.shape
#     if point_0[0] > point_1[0]:
#         warnings.warn("Invalid x_axis slicing, \
#             point_0: {} and point_1: {}".format(point_0, point_1))
#     if point_0[1] > point_1[1]:
#         warnings.warn("Invalid y_axis slicing, \
#             point_0: {} and point_1: {}".format(point_0, point_1))
#     if point_0[0] < 0 or point_0[1] < 0 or point_0[0] > shape[0] or point_0[1] > shape[1]:
#         x0, y0 = point_0
#         x0 = max(0, x0)
#         y0 = max(0, y0)
#         x0 = min(shape[0], x0)
#         y0 = min(shape[1], y0)
#         warnings.warn("Invalid point_0: {}, corrected to {}".format(point_0, (x0, y0)))
#         point_0 = (x0, y0)
#     if point_1[0] < 0 or point_1[1] < 0 or point_1[0] > shape[0] or point_1[1] > shape[1]:
#         x1, y1 = point_1
#         x1 = max(0, x1)
#         y1 = max(0, y1)
#         x1 = min(shape[0], x1)
#         y1 = min(shape[1], y1)
#         warnings.warn("Invalid point_1: {}, corrected to {}".format(point_1, (x1, y1)))
#         point_1 = (x1, y1)
#     return point_0, point_1

def check_borders_correct(array_np, point):
    shape = array_np.shape
    if point[0] < 0 or point[1] < 0 or point[0] > shape[0] or point[1] > shape[1]:
        x, y = point
        x = max(0, x)
        y = max(0, y)
        x = min(shape[0], x)
        y = min(shape[1], y)
        warnings.warn("Invalid point: {}, corrected to {}".format(point, (x, y)))
        point = (x, y)
    return point


def pj_slice(array_np, point_0, point_1=None):
    """
    Allows to slice numpy array's given one point or 
    two points.
    Args:
        array_np : Numpy array to slice
        point_0 : A tuple, or tuple like object of size 2 
                  with integers.
        point_1 : None (default) or a tuple, or tuple like 
                  object of size 2 with integers.
    Returns:
        If point_1 is None, returns array_np evaluated in point_0,
        else returns a slice of array_np between point_0 and point_1.
    """
    x_0, y_0 = check_borders_correct(array_np, point_0)
    if point_1 is None:
        result = array_np[x_0, y_0]
    else:
        x_1, y_1 = check_borders_correct(array_np, point_1)
        if x_0 > x_1:
            warnings.warn("Invalid x_axis slicing, \
                point_0: {} and point_1: {}".format(point_0, point_1))
        if y_0 > y_1:
            warnings.warn("Invalid y_axis slicing, \
                point_0: {} and point_1: {}".format(point_0, point_1))
        result = array_np[x_0:x_1, y_0:y_1]
    return result

def mask_percentage(mask, point, radius, mask_tolerance=0.5):
    """
    Given a binary image and a point and a radius -> sub_img
    Computes a score to know how much of the  sub_img is covered
    by tissue region. Given a tolerance threshold this will return
    a boolean.
    tolerance is mask_tolerance
    tolerance of 1 means that the entire image is in the mask area.
    tolerance of 0.1 means that the image has to overlap at least at 10%
              with the mask.
    Args:
        mask : Binary numpy array, where positive pixels correspond to tissue area and
               negative pixels to background areas in the slide.
        point : A tuple like object of size 2 
                with integers.
        radius : None (default) or a tuple, or tuple like 
                 object of size 2 with integers.
        tolerance : A float between 0 and 1. By default 0.5.
    Returns:
        A boolean. If True, keep, else discard
    """
    sub_mask = pj_slice(mask, point - radius, point + radius + 1)
    score = sub_mask.sum() / (sub_mask.shape[0] * sub_mask.shape[1])
    accepted = score > mask_tolerance
    return accepted



# -*- coding: utf-8 -*-
"""
Created on Tue May 03 16:32:53 2016

@author: Peter
A collection of functions that can be used to generate a mask for a given tissue.
These functions were fine tuned on a personnal data. They can be used at your own risk.
"""
import numpy as np

from skimage.morphology import disk, opening, closing, dilation, remove_small_objects
from skimage.filters import threshold_otsu
from skimage.measure import label
from scipy.ndimage.morphology import binary_fill_holes



def opening_(image, disk_size=7):
    # Computes an opening of size disk.
    inter_image = image.copy()
    inter_image = opening(inter_image, disk(disk_size))
    return inter_image


def tissue_thresh(image, thresh=None):
    # tissue is in black... Tries to find the best background
    if thresh is None:
        thresh = threshold_otsu(image)
    binary = image > thresh
    if binary.dtype == 'bool':
        binary = binary + 0
    return binary


def fill_hole(bin_image, invert=False):
    # files holes
    values = np.unique(bin_image)
    if len(values) > 2:
        print("Not binary image")
        return []
    background = min(values)
    bin_image -= background
    bin_image[bin_image > 0] = 1
    if invert:
        bin_image -= 1
        bin_image[bin_image < 0] = 1
    result = np.copy(bin_image)
    binary_fill_holes(bin_image, output=result)
    return result


def remove_border(image_bin, border=15):
    # removes borders
    neg_border = border * -1
    result = np.copy(image_bin)
    result[:, :] = 0
    result[border:neg_border, border:neg_border] = image_bin[
        border:neg_border, border:neg_border]
    return result


def remove_isolated_points(binary_image, thresh=100):
    # removing tiny areas...
    # pdb.set_trace()
    lbl = label(binary_image)
    lbl = remove_small_objects(lbl, thresh)
    binary = (lbl > 0).astype(int)
    return binary


def find_ticket(rgb_image, _3tuple=(80, 80, 80)):
    # Find the "black ticket on the images"
    temp_image_3 = np.copy(rgb_image)
    temp_image_3[:, :, :] = 0
    for i in range(3):
        temp_image_1 = np.zeros(shape=rgb_image.shape[0:2])
        temp_image_1[np.where(rgb_image[:, :, i] < _3tuple[i])] = 1
        temp_image_3[:, :, i] = temp_image_1

    temp_resultat = temp_image_3.sum(axis=2)

    temp_resultat[temp_resultat > 2] = 3
    temp_resultat[temp_resultat < 3] = 0
    temp_resultat[temp_resultat == 3] = 1
    
    #temp_resultat = Filling_holes_2(temp_resultat)
    temp_resultat = closing(temp_resultat, disk(20))
    temp_resultat = opening_(temp_resultat, 20)
    temp_resultat = remove_border(temp_resultat)
    return temp_resultat


def preprocessing(image, thresh=200, invert=True):
    inter = opening_(image)
    inter = tissue_thresh(inter, thresh)
    inter = fill_hole(inter, invert=invert)
    inter = remove_border(inter)
    res = remove_isolated_points(inter)
    return res

def combining(numpy_array):
    res = np.sum(numpy_array, axis=2)
    res[res > 0] = 1
    return res

def roi_binary_mask2(sample, size=5, ticket=(80, 80, 80)):
    #  very slow function at resolution 4
    preproc_res = np.copy(sample)

    for i in range(3):  # RGB
        # this one is painfully slow..
        preproc_res[:, :, i] = preprocessing(sample[:, :, i])
    res = combining(preproc_res)
    ticket = find_ticket(sample, ticket)
    res = res - ticket
    res[res > 0] = 1
    res[res < 0] = 0
    res = opening_(res, size)
    return res

def roi_binary_mask(sample, size=5):
    val = threshold_otsu(sample[:, :, 0])
    mask = (sample[:, :, 0] < val).astype(int)
    mask = dilation(mask, disk(size))
    mask = fill_hole(mask)
    mask = remove_isolated_points(mask)
    return mask



import ez_setup
ez_setup.use_setuptools()

from setuptools import setup, find_packages

exec(open('useful_wsi/version.py').read()) # loads __version__

setup(name='useful_wsi',
      version=__version__,
      author='Peter',
      description='',
      long_description=open('README.rst').read(),
      license='see LICENSE.txt',
      keywords="",
      packages=find_packages(exclude='docs'))



"""

.. moduleauthor:: Naylor Peter <peter.naylor@mines-paristech.fr>

"""
from .tissue_segmentation import roi_binary_mask, roi_binary_mask2
from .patch_sampling import patch_sampling
from .patch_sampling import sample_patch_from_wsi, random_wsi_sampling
from .utils import *
from .visu_utils import visualise_cut, PLOT_ARGS
from .version import __version__



# -*- coding: utf-8 -*-
"""
Created on Tue May 03 16:32:53 2016

@author: Peter
"""

import matplotlib.patches as patches
import matplotlib.pyplot as plt

from .utils import open_image, get_whole_image, get_x_y_from_0, get_size

PLOT_ARGS = {'color': 'red', 'size': (12, 12), 'title': ""}

def visualise_cut(slide, list_pos, res_to_view=None, plot_args=PLOT_ARGS):
    """
    Plots the patches you are going to extract from the slide. So that they
    appear as red boxes on the lower resolution of the slide.
    Args:
        slide : str or openslide object.
        list_pos : list of parameters to extract tiles from slide.
        res_to_view : integer (default: None) resolution at which to
                      view the patch extraction.
        plot_args : dictionnary for any plotting argument.
    """
    slide = open_image(slide)
    if res_to_view is None:
        res_to_view = slide.level_count - 1
    elif res_to_view > slide.level_count - 1:
        print(" level ask is too low... It was setted accordingly")
        res_to_view = slide.level_count - 1
    whole_slide = get_whole_image(slide, res_to_view, numpy=True)
    fig = plt.figure(figsize=plot_args['size'])
    axes = fig.add_subplot(111, aspect='equal')
    axes.imshow(whole_slide)
    for para in list_pos:
        top_left_x, top_left_y = get_x_y_from_0(slide, (para[0], para[1]), res_to_view)
        width, height = get_size(slide, (para[2], para[3]), para[4], res_to_view)
        plot_seed = (top_left_x, top_left_y)
        patch = patches.Rectangle(plot_seed, width, height,
                                  fill=False, edgecolor=plot_args['color'])
        axes.add_patch(patch)
    axes.set_title(plot_args['title'], size=20)
    axes.axis('off')
    plt.show()
